-- by mor2000
-- v1.0
-- PlaceableHeaps

source(g_currentModDirectory.."HeapTipTrigger.lua");
source(g_currentModDirectory.."HeapShovelFillTrigger.lua");

PlaceableHeap = {};
InitObjectClass(PlaceableHeap, "PlaceableHeap");

function PlaceableHeap:new(isServer, isClient, mt)
	--print("new Heap");
	if PlaceableHeap_mt == nil then
		PlaceableHeap_mt = Class(PlaceableHeap, Placeable);
	end;
	
	if mt == nil then
		mt = PlaceableHeap_mt;
	end;

    self = Placeable:new(isServer, isClient, mt);
    registerObjectClassName(self, "PlaceableHeap");
	
	self.isServer=isServer;
	self.isClient=isClient;
	self.mt=mt;

	self.nodeId=0;

	self.update=PlaceableHeap.update;

	self.setFillLevel=PlaceableHeap.setFillLevel;
	self.getFillLevel=PlaceableHeap.getFillLevel;

	self.getAcceptedFillTypes=PlaceableHeap.getAcceptedFillTypes;
	self.setFillType=PlaceableHeap.setFillType;
	self.getFillType=PlaceableHeap.getFillType;
	
	self.getAllowShovelFillType = PlaceableHeap.getAllowShovelFillType;
	self.addShovelFillLevel = PlaceableHeap.addShovelFillLevel;
	
	self.displayTriggerCallback = PlaceableHeap.displayTriggerCallback;
	
	self.tipTrigger=0;
	self.shovelTrigger=nil;

	self.heapTipTriggerDirtyFlag = self:getNextDirtyFlag();
	
	self.fillLevel=0;
	self.acceptedFillTypes={};
	self.fillType=0;
	
	self.heapPlanes={};
	self.activePlaneId=0;
	self.activePlaneShovelTriggerId=0;
	self.activePlaneShovelTriggerOn=false;
	self.activePlaneX=0;
	self.activePlaneY=nil;
	self.activePlaneZ=0;
	self.moveMaxAmount=700000;
	self.moveMaxY=0;
	self.moveMinY=-3.75;
	
	self.getHemisphereRadiusFromVolume=PlaceableHeap.getHemisphereRadiusFromVolume;
	self.hemisphereRadiusMaxAmount=0;

	self.forceMovePlane = PlaceableHeap.forceMovePlane;
	self.activePlaneRepostioned=false;
	
	self.displayTriggerId=0;
	self.showFillLevelDisplay = false;
	
	--print("Heap created");
	return self;
end;

function PlaceableHeap:delete()
	if self.tipTrigger~=nil and self.tipTrigger~=0 then
		self.tipTrigger:delete();
	end;

	if(self.shovelTrigger~=nil and self.shovelTrigger~=0)then
		self.shovelTrigger:delete();
	end;

	if(self.displayTriggerId~=0)then
		removeTrigger(self.displayTriggerId);
	end;
	
	unregisterObjectClassName(self);
    PlaceableHeap:superClass().delete(self);
end;

function PlaceableHeap:deleteFinal()
    PlaceableHeap:superClass().deleteFinal(self);
end;

-- Function to show Id-Tree of I3D

--[[
function getChildren(id)
	for i=1,getNumOfChildren(id) do
		child=getChildAt(id, i-1);
		name=Utils.getNoNil(getName(child),"na");
		--print("found Child of "..id.." with Node "..child.." and Name "..name);
		getChildren(child);
	end;
end;
--]]

function PlaceableHeap:load(xmlFilename, x,y,z, rx,ry,rz, moveMode, initRandom)
	--print("loading Heap");
	if not PlaceableHeap:superClass().load(self, xmlFilename, x,y,z, rx,ry,rz, moveMode, initRandom) then
        return false;
    end;
	--print("loading Heap successful");
	
	--print("Filename: "..xmlFilename..", moveMode: "..moveMode);

	self.xml = xmlFilename;
	self.pos = {x,y,z};
	self.rot = {rx,ry,rz};
	
	if not moveMode then
		local xmlFile = loadXMLFile("TempXML", xmlFilename);
		
		-- set accepted FillTypes
		local fillTypesString=Utils.getNoNil(getXMLString(xmlFile, "placeable.FillTypes#names"),"");
		for k,v in pairs(Utils.splitString(" ",fillTypesString)) do
			local desc2 = Fillable.fillTypeNameToDesc[v];
			if type(desc2)=="table" then
				--print("index for "..v..": "..desc2.index);
				self.acceptedFillTypes[desc2.index] = true;
			else
				--print("index for "..v.." not found");
			end;
		end;
		
		

		--print("attempt to add Heap Trigger"); -- 1 tipTrigger expected
		--getChildren(self.nodeId); -- see structure of Ids
		local TriggerGroupId=Utils.getNoNil(getChild(self.nodeId,"Trigger"),0);
		if TriggerGroupId~=0 then
			local tipTriggerId=Utils.getNoNil(getChild(TriggerGroupId,"tipTrigger"),0);
			if tipTriggerId ~= 0 then
				--print("tipTrigger found at "..tipTriggerId);
				self.tipTrigger=HeapTipTrigger:onCreate(tipTriggerId,self);
				--self.tipTriggers[1]:setAcceptedFillTypes(self.acceptedFillTypes);
				--print("new Heap Trigger "..tipTriggerId.." added");
			end;
			
			local shovelTriggerId=Utils.getNoNil(getChild(TriggerGroupId,"shovelTrigger"),0);
			if shovelTriggerId ~= 0 then
				--print("tipTrigger found at "..tipTriggerId);
				self.shovelTriggerId=shovelTriggerId;
				--self.tipTriggers[1]:setAcceptedFillTypes(self.acceptedFillTypes);
				--print("new Heap Trigger "..tipTriggerId.." added");
			end;
			
			local displayTriggerId=Utils.getNoNil(getChild(TriggerGroupId,"displayTrigger"),0);
			if displayTriggerId ~= 0 then
				--print("displayTrigger found at "..displayTriggerId);
				self.displayTriggerId=displayTriggerId;
				addTrigger(self.displayTriggerId, "displayTriggerCallback", self);
				--[[	<Shape name="displayTrigger" translation="0 0 0" scale="1 1 3" static="true" trigger="true" collisionMask="15736832" nodeId="46" materialIds="9" shapeId="1" nonRenderable="true"/>
				--]]
			end;
		end;

		--print("attempt to add Heap Planes");
		local heapPlanesId = Utils.getNoNil(getChild(self.nodeId,"HeapPlanes"),0);
		if(heapPlanesId~=0)then
			self.moveMaxAmount=Utils.getNoNil(getUserAttribute(heapPlanesId, "moveMaxAmount"), self.moveMaxAmount);
			self.moveMaxY=Utils.getNoNil(getUserAttribute(heapPlanesId, "moveMaxY"), self.moveMaxY);
			self.moveMinY=Utils.getNoNil(getUserAttribute(heapPlanesId, "moveMinY"), self.moveMinY);
			for i=1,getNumOfChildren(heapPlanesId) do
				local heapPlaneId = getChildAt(heapPlanesId, i-1);
				if(heapPlaneId~=0)then
					local fillTypeString=Utils.getNoNil(getUserAttribute(heapPlaneId, "fillType"), 0);
					local desc2 = Fillable.fillTypeNameToDesc[fillTypeString];
					if type(desc2)=="table" then
						--print("Heap for Filltype "..fillTypeString.."("..tostring(desc2.index)..") found");
						self.heapPlanes[desc2.index]= heapPlaneId;
						--print("Heap Plane "..self.heapPlanes[desc2.index].." of Filltyp "..desc2.index.." added");
					end;
				end;
			end;
		end;

		delete(xmlFile);
		
	end;

	self.showFillLevelDisplay=false;
	
	--print("loading Heap finished");
	return true;
end;


function PlaceableHeap:update(dt)
	--print(self:getFillLevel());
	if self.showFillLevelDisplay then
		g_currentMission:addExtraPrintText(g_i18n:getText("fill_level").." "..math.floor(self:getFillLevel()));
	end;
	
	--self.timeSinceLastUpdate=self.timeSinceLastUpdate+dt;
	if(self.activePlaneRepostioned)then
		if(self.activePlaneId~=0)then
			setTranslation(self.activePlaneId,self.activePlaneX,self.activePlaneY,self.activePlaneZ);
		end;
		setTranslation(self.shovelTriggerId,self.activePlaneX,self.activePlaneY+0.5,self.activePlaneZ);
		self.activePlaneRepostioned=false;
	end;
	--print("Timediff is "..dt);
	
end;

function PlaceableHeap:getAllowShovelFillType(fillType)
	--print("Ask to fill from Shovel of Type "..tostring(fillType));
	if(self.fillType==0)then
		return self.acceptedFillTypes[fillType];
	else
		return self.fillType==fillType;
	end;
end;

function PlaceableHeap:addShovelFillLevel(shovel, fillLevelDelta, fillType)
	--print("addShovelFillLevel with "..tostring(fillLevelDelta));
	if self:getAllowShovelFillType(fillType) then
		self:setFillType(fillType);
		local oldFillLevel = self:getFillLevel();
		--print("Old FillLevel: "..tostring(oldFillLevel)..", new one: "..tostring(oldFillLevel+fillLevelDelta));
		self:setFillLevel(oldFillLevel+fillLevelDelta);
		fillLevelDelta = self:getFillLevel() - oldFillLevel;
		return fillLevelDelta;
	end;
	return 0;
end;

function PlaceableHeap:updateMoving(delta)
end;

function PlaceableHeap:getSaveAttributesAndNodes(nodeIdent)
	local x,y,z = unpack(self.pos);
	local rx,ry,rz = unpack(self.rot);
	local fillLevel = tostring(self:getFillLevel());
	local fillType = tostring(self:getFillType());

	local attributes = 'filename="'..Utils.encodeToHTML(Utils.convertToNetworkFilename(self.xml))..'" position="'..x..' '..y..' '..z..'" rotation="'..rx..' '..ry..' '..rz..'" fillLevel="'..fillLevel..'" fillType="'..fillType..'"';
	
	local nodes = "";

	return attributes,nodes;
end;

function PlaceableHeap:loadFromAttributesAndNodes(xmlFile, Key)
	--print("loading Heap Attributes");
	local xmlFilename =  Utils.convertFromNetworkFilename(getXMLString(xmlFile, Key.."#filename"));
	local x,y,z = Utils.getVectorFromString(getXMLString(xmlFile, Key.."#position"));
	local rx,ry,rz = Utils.getVectorFromString(getXMLString(xmlFile, Key.."#rotation"));
	
	if self:load(xmlFilename, x,y,z, rx,ry,rz) then
		local fillLevel = Utils.getNoNil(getXMLFloat(xmlFile, Key.."#fillLevel"),0);
		local fillType = getXMLFloat(xmlFile, Key.."#fillType");
		if(fillType=="nil")then fillType=nil; end;
		--print("Heap with fillTypes "..tostring(fillType).." and fillLevel "..fillLevel.." loaded");
		self:setFillLevel(0);
		self:setFillType(0);
		--print("Attempt to fill Heap with "..fillLevel);
		self:setFillLevel(fillLevel);
		self:setFillType(fillType);
	end;

	return true;
end;

function PlaceableHeap:getFillLevel()
	--print ("Heap getFillLevel called");
	return self.fillLevel;
end;

function PlaceableHeap:setFillLevel(newFillLevel)
	local oldFillLevel = self.fillLevel;
	self.fillLevel = newFillLevel;
	
	--print("Old FillLevel: "..tostring(oldFillLevel)..", new one: "..tostring(self.fillLevel));
	
	--print("Heap FillLevel set to "..newFillLevel.." = "..tostring(self.fillLevel));

	if self.fillLevel < 0.001 then
		self.fillLevel = 0;
		self:setFillType(0);
		self.activePlaneY=self.moveMinY-1;
		self.activePlaneRepostioned=true;
	else
		if self.fillLevel ~= oldFillLevel then
			self:forceMovePlane();
		end;
	end;
end;

function PlaceableHeap:getHemisphereRadiusFromVolume(volume)
	if volume==0 then
		return 0;
	else
		return (volume/(4/6*math.pi))^(1/3);
	end;
end;

function PlaceableHeap:forceMovePlane()
	if(self.activePlaneId~=0)then
		--print("attempt to move Plane");
		if(self.hemisphereRadiusMaxAmount==0)then
			self.hemisphereRadiusMaxAmount=self:getHemisphereRadiusFromVolume(self.moveMaxAmount);
		end;
		local amount=math.min(self.moveMaxAmount,Utils.getNoNil(self:getFillLevel(),0));
		local ratio=self:getHemisphereRadiusFromVolume(amount)/self.hemisphereRadiusMaxAmount;
		--print("Ratio: "..tostring(ratio));
		--print('FillLevel at '..tostring(self.moveMaxAmount).." has "..tostring(self.radiusMoveMaxAmount))
		--print('FillLevel at '..tostring(amount).." has "..tostring(radiusAmount));
		self.activePlaneY=(self.moveMaxY-self.moveMinY)*ratio+self.moveMinY;
		self.activePlaneRepostioned=true;
		--print("Set Y Translation to "..tostring(newY));
		--setTranslation(self.activePlaneId, self.activePlaneX, newY, self.activePlaneY);
	end;
end;

function PlaceableHeap:setFillType(fillType)
	-- change plane texture
	if(fillType==0)then
		-- hide plane of old filltype
		if(self.shovelTrigger~=nil and self.shovelTrigger~=0)then
			self.shovelTrigger:delete();
			--print("Removed Shovel Trigger");
			self.activePlaneShovelTriggerOn=false;
		end;
		--print("Hiding Plane of old FillType");
		if self.activePlaneId~=nil and self.activePlaneId~=0 then
			setVisibility(self.activePlaneId,false);
		end;
		self.activePlaneY=self.moveMinY-1;
		self.activePlaneRepostioned=true;
	end;
	
	if(self.fillType~=fillType and fillType~=0)then
		--print("Heap has new Filltype: "..tostring(fillType));
		if self.shovelTriggerId~=nil then
			-- show plane of new fill type
			self.shovelTrigger = HeapShovelFillTrigger:onCreate(self.shovelTriggerId,self,fillType);
			--print("Added Shovel Trigger with Id "..self.shovelTriggerId);
			--print("Showing Plane of new FillType: "..tostring(self.heapPlanes[fillType].id).." with "..tostring(fillType));
			self.activePlaneId=self.heapPlanes[fillType];
			setVisibility(self.activePlaneId,true);
			self:forceMovePlane();
		end;
	end;

	-- change filltype
	if(fillType~=nil and fillType~=0)then
		if self.acceptedFillTypes[fillType]==true then
			self.fillType=fillType;
		end;
	else
		self.fillType=0;
	end;
end;

function PlaceableHeap:getFillType()
	return self.fillType;
end;

function PlaceableHeap:getAcceptedFillTypes()
	return self.acceptedFillTypes;
end;

function PlaceableHeap:displayTriggerCallback(triggerId, otherActorId, onEnter, onLeave, onStay, otherShapeId)
-- print("Display Trigger called");
	if onEnter==true then
	local otherActor=g_currentMission.nodeToVehicle[otherActorId];
	if(otherActor~=nil and otherActor:getOwner())then
		self.showFillLevelDisplay=true;
	end;
else
	self.showFillLevelDisplay=false;
end;

end;

registerPlaceableType("PlaceableHeap", PlaceableHeap);